
#include "RFVoxelOctree.h"

using namespace RFAxl;

RFVoxelOctreeNode::RFVoxelOctreeNode() :
m_eNodeType(NT_INTERNAL),
m_fDistanceToSurface(0.0f)
{
	for (uint i = 0; i < 8; i++)
	{
		m_pkChildren[i] = nullptr;
	}
}

RFVoxelOctreeNode::~RFVoxelOctreeNode()
{
	for (unsigned int i = 0; i < 8; i++)
	{
		delete m_pkChildren[i];
	}
}


RFVoxelOctree::RFVoxelOctree() : 
m_pkRoot(nullptr),
m_fVoxelSize(1.0f), 
m_kDataDimension(0), 
m_kOctreeOrigin(0.0f),
m_uiOctreeDim(0), 
m_uiOctreeDepth(0), 
m_uiNumberOfNodes(0), 
m_uiNumberOfLeafs(0), 
m_uiNumberOfEmptyLeafs(0), 
m_uiNumberOfInternalNodes(0)
{

}

RFVoxelOctree::~RFVoxelOctree()
{
	delete m_pkRoot;
}

bool RFVoxelOctree::IsValid() const
{
	return m_pkRoot != nullptr;
}

uint RFVoxelOctree::GetDepth() const
{
	return static_cast<uint>(std::log2(m_uiOctreeDim));
}


float RFVoxelOctree::GetVoxelSize() const
{
	return m_fVoxelSize;
}

RFMath::RFVector3f RFVoxelOctree::GetVoxelCornerPosition() const
{
	return m_pkRoot->m_kWorldAabb.kMin;
}
